#include "cm.h"

#include "geometry.h"
#include "shader.h"
#include "wm.h"
#include "util.h"

#ifndef SHADERS_PATH
#define SHADERS_PATH                                "assets/shaders"
#endif

#define MAIN_SHADER_NAME                            "main"

#define XCOMPMGR_PROPERTY                           "xcompmgr"

// Taken from https://github.com/obiwac/x-compositing-wm/blob/main/src/cwm.h
typedef GLXContext (*glXCreateContextAttribsARB_t)  (Display*, GLXFBConfig, GLXContext, Bool, const int*);
typedef void (*glXBindTexImageEXT_t)                (Display*, GLXDrawable, int, const int*);
typedef void (*glXReleaseTexImageEXT_t)             (Display*, GLXDrawable, int);
typedef void (*glXSwapIntervalEXT_t)                (Display*, GLXDrawable, int);

typedef struct window_vertices_t
{
    float vertices[4 * 2];
} window_vertices_t;

static const char * MAIN_SHADER_UNIFORM_NAMES[] = {
    // "screen_width",
    // "screen_height",

    // "window_x",
    // "window_y",

    // "window_width",
    // "window_height",

    "texture_sampler",

    "global_transform"
};

typedef struct main_shader_t
{
    GLuint program;

    struct
    {
        // int screen_width;
        // int screen_height;

        // int window_x;
        // int window_y;

        // int window_width;
        // int window_height;

        int texture_sampler;

        int global_transform;
    } uniforms;
} main_shader_t;

typedef struct window_extra_t
{
    bool drawable;

	Pixmap xpixmap;
	GLXPixmap pixmap;
} window_extra_t;

typedef struct cm_t
{
    int screen;

    Display* display;
    Window root_window;

    XWindowAttributes root_window_attributes, 
        output_window_attributes;

    Window check_window, overlay_window,
        output_window;

    Atom net_wm_cm_atom;

    XVisualInfo * default_visual;
    GLXFBConfig * fb_config;
    int fb_config_count;
    GLXContext glx_context;
    
    glXCreateContextAttribsARB_t glXCreateContextAttribsARB;

	glXBindTexImageEXT_t glXBindTexImageEXT;
	glXReleaseTexImageEXT_t glXReleaseTexImageEXT;
    glXSwapIntervalEXT_t glXSwapIntervalEXT;

    main_shader_t main_shader;

    bool allocated_buffer;
    unsigned int vbo, vao, ebo;
} cm_t;

static cm_t cm = {0};

static const int 
DEFAULT_GLX_VISUAL_ATTRIBUTES[] = {
    GLX_RGBA,                               GLX_DOUBLEBUFFER,
    GLX_SAMPLE_BUFFERS,                     1,
    GLX_SAMPLES,                            4,
    GLX_RED_SIZE,                           8,
    GLX_GREEN_SIZE,                         8,
    GLX_BLUE_SIZE,                          8,
    GLX_ALPHA_SIZE,                         8,
    GLX_DEPTH_SIZE,                         16, 
    
    0
};

static const int 
DEFAULT_GLX_FB_ATTRIBUTES[] = {
    GLX_BIND_TO_TEXTURE_RGBA_EXT,           1,
    GLX_BIND_TO_TEXTURE_TARGETS_EXT,        GLX_TEXTURE_2D_BIT_EXT,
    GLX_RENDER_TYPE,                        GLX_RGBA_BIT,
    GLX_DRAWABLE_TYPE,                      GLX_PIXMAP_BIT,
    GLX_X_VISUAL_TYPE,                      GLX_TRUE_COLOR,
    GLX_X_RENDERABLE, 1,
    GLX_FRAMEBUFFER_SRGB_CAPABLE_EXT,       (GLint) GLX_DONT_CARE,
    GLX_BUFFER_SIZE,                        32,
    GLX_SAMPLE_BUFFERS,                     1,
    GLX_SAMPLES,                            4,
    GLX_DOUBLEBUFFER,                       1,
    GLX_RED_SIZE,                           8,
    GLX_GREEN_SIZE,                         8,
    GLX_BLUE_SIZE,                          8,
    GLX_ALPHA_SIZE,                         8,
    GLX_STENCIL_SIZE,                       0,
    GLX_DEPTH_SIZE,                         16,
    
    0
};

static const int 
DEFAULT_GLX_VERSION_ATTRIBUTES[] = {
    GLX_CONTEXT_MAJOR_VERSION_ARB,          3,
    GLX_CONTEXT_MINOR_VERSION_ARB,          3,
    GLX_CONTEXT_FLAGS_ARB,                  GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB, 
    
    0
};

static const unsigned int 
WINDOW_INDICES[] = {
    0, 1, 3,
    1, 2, 3,
};


static const float
WINDOW_VERTICES[] = {
    // Top right
    1.0, -1.0, 0.0f,

    // Botom right
    1.0, 1.0, 0.0f,

    // Bottom left
    -1.0, 1.0, 0.0f,

    // Top left
    -1.0, -1.0, 0.0f
};

static inline bool
should_render_window( window_node_t * node )
{
    return (
        node->window != cm.check_window     &&
        node->window != cm.overlay_window   &&
        node->window != cm.output_window    &&
        node->window != cm.root_window      &&
        node->attributes.width > 1          &&
        node->attributes.height > 1
    );
}

static void
free_window_pixmap(window_extra_t *extra)
{
	if (extra->xpixmap)
		XFreePixmap(cm.display, extra->xpixmap);

	if (extra->pixmap)
		glXDestroyPixmap(cm.display, extra->pixmap);
    
    extra->xpixmap = extra->pixmap = 0;
}

static void
create_window_pixmap( window_node_t * window_node )
{
    XVisualInfo* visual = NULL;
    int format = 0, has_alpha = 0, visual_depth = 0;
    window_extra_t * extra = NULL;
    GLXFBConfig config = {0};

    extra = (window_extra_t *)window_node->extra;

    XGetWindowAttributes(cm.display, window_node->window, &window_node->attributes);

    for (int i = 0; i < cm.fb_config_count; i++) {
        config = cm.fb_config[i];

        glXGetFBConfigAttrib(cm.display, config, GLX_BIND_TO_TEXTURE_RGBA_EXT, &has_alpha);

        visual = glXGetVisualFromFBConfig(cm.display, config);
        visual_depth = visual->depth;
        free(visual);

        if (window_node->attributes.depth != visual_depth)
            continue;

        format = has_alpha ? GLX_TEXTURE_FORMAT_RGBA_EXT : GLX_TEXTURE_FORMAT_RGB_EXT;
        break;
    }
    
    int pixmap_attributes[] = {
        GLX_TEXTURE_TARGET_EXT, GLX_TEXTURE_2D_EXT,
        GLX_TEXTURE_FORMAT_EXT, format, 
        
        0
    };
    extra->xpixmap = XCompositeNameWindowPixmap(cm.display, window_node->window);
    extra->pixmap = glXCreatePixmap(cm.display, config, extra->xpixmap, (const int *)pixmap_attributes);
}

static inline window_extra_t *
create_window_extra( window_node_t * node )
{
    window_extra_t * extra = NULL;

    extra = gcalloc(1, sizeof(window_extra_t));
    node->extra = extra;

    return extra;
}

static window_extra_t *
create_window_extra_if_none( window_node_t * node )
{
    if(!node->extra)
        node->extra = create_window_extra(node);

    return (window_extra_t *)node->extra;
}

static char *
get_shader_path(const char * filename)
{
    size_t size = 0;
    char * path = NULL;

    size = arrlen(SHADERS_PATH) + strlen(filename) + 2;
    path = gcalloc(size, sizeof(char));

    snprintf(path, size, "%s/%s", SHADERS_PATH, filename);

    return path;
}

static int
compile_shader(const char * name, const char *vertex_shader_filename, 
    const char *fragment_shader_filename, const char *uniform_names[], 
    int *uniforms, size_t uniform_len)
{
    int ret = 0;
    char * path = NULL;
    char *vertex_source = NULL, *fragment_source = NULL;

    dbg("compiling \"%s\" shader...", name);

    path = get_shader_path(vertex_shader_filename);
    if((ret = read_file(path, &vertex_source, NULL)))
    {
        err("cannot open %s: %s", path, strerror(ret));
        return ret;
    }
    gfree(path);

    path = get_shader_path(fragment_shader_filename);
    if((ret = read_file(path, &fragment_source, NULL)))
    {
        err("cannot open %s: %s", path, strerror(ret));
        return ret;
    }
    gfree(path);

    ret = shader_create(&cm.main_shader.program, vertex_source, fragment_source,
        uniform_names, uniforms, uniform_len);
    if(ret)
        return ret;
    
    dbg("done compiling \"%s\" shader", name);
    
    return 0;
}

static int
compile_shaders( void )
{
    int ret = 0;

    ret = compile_shader(MAIN_SHADER_NAME, MAIN_SHADER_NAME ".vs", MAIN_SHADER_NAME ".fs",
        MAIN_SHADER_UNIFORM_NAMES, (int *)&cm.main_shader.uniforms, 
        arrlen(MAIN_SHADER_UNIFORM_NAMES));
    if(ret)
        return ret;

    return 0;
}

static int 
bind_window_texture( window_node_t * window_node )
{
    window_extra_t * extra = NULL;

    extra = window_node->extra;
    if(!extra || !extra->drawable)
        return 0;

    if(!extra->pixmap)
        create_window_pixmap(window_node);
    cm.glXBindTexImageEXT(cm.display, extra->pixmap, GLX_FRONT_LEFT_EXT, NULL);

    return 0;
}

static int
unbind_window_texture( window_node_t * window_node )
{
    window_extra_t * extra = NULL;
    
    extra = window_node->extra;
    
    if(extra->pixmap)
        cm.glXReleaseTexImageEXT(cm.display, extra->pixmap, GLX_FRONT_LEFT_EXT);
    
    return 0;
}

int
cm_swap_buffers( void )
{
    glXSwapBuffers(cm.display, cm.output_window);
    return 0;
}

static int
render_window( window_node_t * window_node )
{
    int ret = 0;
    matrix_t ma, mb, transform;
    float height = 0.0f, width = 0.0f, x = 0.0f, y = 0.0f,
        output_width = 0.0f, output_height = 0.0f;
    window_extra_t * extra = NULL;

    extra = create_window_extra_if_none(window_node);

    XGetWindowAttributes(cm.display, cm.output_window, &cm.output_window_attributes);
    XGetWindowAttributes(cm.display, window_node->window, &window_node->attributes);

    x = (float)window_node->attributes.x;
    y = (float)window_node->attributes.y;

    width = (float)window_node->attributes.width;
    height = (float)window_node->attributes.height;

    output_width = (float)cm.output_window_attributes.width;
    output_height = (float)cm.output_window_attributes.height;

    if(window_node->attributes.width % 2) x += .5f;
    if(window_node->attributes.height % 2) y += .5f;
    
    y = -y;

    // Setup unifroms for main shader
    glUseProgram(cm.main_shader.program);

	glUniform1i(cm.main_shader.uniforms.texture_sampler, 0);

    make_scaling_matrix(width/output_width, height/output_height, 0.0f, ma);
    make_translation_matrix((x - width/2.0f)/output_width, (y + height/2.0f)/output_height, 0.0f, mb);
    mxm(mb, ma, transform);

    glUniformMatrix4fv(cm.main_shader.uniforms.global_transform, 1, GL_FALSE, &transform[0][0]);

    // Bind texture
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glActiveTexture(GL_TEXTURE0);

    bind_window_texture(window_node);

    glBindVertexArray(cm.vao);
    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);

    unbind_window_texture(window_node);

    return ret;
}

int
cm_render_windows(void)
{
    window_node_t * node = NULL;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glBindVertexArray(cm.vao);

    glBindBuffer(GL_ARRAY_BUFFER, cm.vbo);
    glBufferData(GL_ARRAY_BUFFER, sizeof(WINDOW_VERTICES), WINDOW_VERTICES, GL_STATIC_DRAW);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cm.ebo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(WINDOW_INDICES), WINDOW_INDICES, GL_STATIC_DRAW);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    glBindBuffer(GL_ARRAY_BUFFER, 0); 
    glBindVertexArray(0); 

    // XGrabServer(cm.display);

    node = wm_get_first_window_node();
    while(node)
    {
        if(should_render_window(node))
            render_window(node);
        node = node->next;
    }

    uint64_t delta = time_ms();
    cm_swap_buffers();
    delta = time_ms() - delta;

    // XUngrabServer(cm.display);

    return 0;
}

static inline void
free_window_extra(window_node_t * node)
{
    window_extra_t * extra = NULL;

    if(extra)
    {
        extra = (window_extra_t *)node->extra;

        free_window_pixmap(extra);
        gfree(extra);
        extra = NULL;
    }
}

static void
on_create_window(window_node_t * node)
{
    if(node->extra)
        free_window_extra(node->extra);

    create_window_extra(node);
    ((window_extra_t *)node->extra)->drawable = true;
}

static void
on_destroy_window(window_node_t * node)
{
    free_window_extra(node);
}

int
cm_init( unsigned int modkey )
{
    int err = 0;
    XserverRegion region = {0};
    wm_callbacks_t callbacks = {0};
    XWindowAttributes window_attrbiutes = {0};
    XSetWindowAttributes default_visual_attributes = {0};
    char net_wm_cm[15] = "_NET_WM_CM_S##";

    clear(&cm);

    err = wm_init(modkey);
    if(err)
        ftl("cannot initialize window manager");

    cm.screen = wm_get_screen();
    cm.display = wm_get_display();
    cm.root_window = wm_get_root_window();

    XGetWindowAttributes(cm.display, cm.root_window, &cm.root_window_attributes);

    // Make it known to other apps that you are a compositing window
    // manager
    cm.check_window = XCreateSimpleWindow(cm.display, cm.root_window, 0, 0, 1, 1, 0, 0, 0);
    Xutf8SetWMProperties(cm.display, cm.check_window, XCOMPMGR_PROPERTY, XCOMPMGR_PROPERTY, 
        NULL, 0, NULL, NULL, NULL);

    snprintf(net_wm_cm, sizeof(net_wm_cm), "_NET_WM_CM_S%d", cm.screen);
    cm.net_wm_cm_atom = XInternAtom(cm.display, net_wm_cm, 0);
    XSetSelectionOwner(cm.display, cm.net_wm_cm_atom, cm.check_window, 0);

    XCompositeRedirectSubwindows(cm.display, cm.root_window, CompositeRedirectManual);

    // We use this so we can make windows transparents to events
	cm.overlay_window = XCompositeGetOverlayWindow(cm.display, cm.root_window);

    region = XFixesCreateRegion(cm.display, NULL, 0);
    XFixesSetWindowShapeRegion(cm.display, cm.overlay_window, ShapeInput, 0, 0, region);
    XFixesDestroyRegion(cm.display, region);

    // Create the window where we will be drawing on
    cm.default_visual = glXChooseVisual(cm.display, cm.screen, (int *)DEFAULT_GLX_VISUAL_ATTRIBUTES);
    if(!cm.default_visual)
        ftl("cannot get default glx visual");
    
    default_visual_attributes = (XSetWindowAttributes){
        .colormap = XCreateColormap(cm.display, cm.root_window, cm.default_visual->visual, AllocNone),
        .border_pixel = 0};

    // Get root window attributes for the output window
    wm_get_window_attributes(cm.root_window, &window_attrbiutes);
    cm.output_window = XCreateWindow(cm.display, cm.root_window, 0, 0, window_attrbiutes.width, 
        window_attrbiutes.height, 0, cm.default_visual->depth, InputOutput, cm.default_visual->visual, 
        CWBorderPixel | CWColormap, &default_visual_attributes);
    
    XReparentWindow(cm.display, cm.output_window, cm.overlay_window, 0, 0);
    XMapRaised(cm.display, cm.output_window);

    // Get the glx frame buffer configurations that match DEFAULT_GLX_FB_ATTRIBUTES
    cm.fb_config = glXChooseFBConfig(cm.display, cm.screen, DEFAULT_GLX_FB_ATTRIBUTES,
        &cm.fb_config_count);
    if(!cm.fb_config)
        ftl("cannot get glx framebuffer configs");

    // Get OpenGL function callbacks
    cm.glXCreateContextAttribsARB = (glXCreateContextAttribsARB_t) glXGetProcAddressARB( (const GLubyte*) "glXCreateContextAttribsARB");
	cm.glXBindTexImageEXT =         (glXBindTexImageEXT_t) glXGetProcAddress((const GLubyte*) "glXBindTexImageEXT");
	cm.glXReleaseTexImageEXT =      (glXReleaseTexImageEXT_t) glXGetProcAddress((const GLubyte*) "glXReleaseTexImageEXT");
    cm.glXSwapIntervalEXT =         (glXSwapIntervalEXT_t) glXGetProcAddress((const GLubyte*) "glXSwapIntervalEXT");

    // Create OpenGL context
    cm.glx_context = cm.glXCreateContextAttribsARB(cm.display, cm.fb_config[0], NULL, 1, DEFAULT_GLX_VERSION_ATTRIBUTES);
    if(!cm.glx_context)
        ftl("cannot create OpenGL context");
    
    // Make that OpenGL context our current context
    glXMakeCurrent(cm.display, cm.output_window, cm.glx_context);

    // Initialize GLEW
	glewExperimental = GL_TRUE;
    if(glewInit() != GLEW_OK)
        ftl("cannot initialize GLEW");

    inf("compiling shaders..");
    err = compile_shaders();
    if(err)
        ftl("cannot compile shaders")        
    inf("done compiling shaders");

    callbacks.on_create_window = on_create_window;
    callbacks.on_destroy_window = on_destroy_window;
    wm_set_callbacks(callbacks);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glGenVertexArrays(1, &cm.vao);
    glGenBuffers(1, &cm.vbo);
    glGenBuffers(1, &cm.ebo);

    cm.allocated_buffer = true;

    glClearColor(0.16015625, 0.15625, 0.15625, 1.);

    return 0;
}

int
cm_destroy( void )
{
    window_node_t *node = NULL, *next_node = NULL;

    node = wm_get_first_window_node();
    while(node)
    {
        next_node = node->next;
        free_window_extra(node);

        node = next_node;
    }

    if(cm.main_shader.program)
        glDeleteProgram(cm.main_shader.program);

    if(cm.allocated_buffer)
    {
        glDeleteVertexArrays(1, &cm.vao);
        glDeleteBuffers(1, &cm.vbo);
        glDeleteBuffers(1, &cm.ebo);
    }

    wm_destroy();
    clear(&cm);

    return 0;
}